/*
/--------------------------------------------------------------------
|
|      $Id: FilterContrast.cpp,v 1.3 2001/01/12 23:30:01 uzadow Exp $
|
|      Copyright (c) 1996-1998 Ulrich von Zadow
|
\--------------------------------------------------------------------
*/

#include "stdpch.h"
#include "FilterContrast.h"
#include "bitmap.h"
#include "anybmp.h"

CFilterContrast::CFilterContrast(double contrast, BYTE offset)
{
  m_contrast = contrast;
  m_offset   = (double) offset-125.0;
}

CFilterContrast::~CFilterContrast()
{

}

void CFilterContrast::Apply(CBmp * pBmpSource, CBmp * pBmpDest) 
{
  // Consider a coordinate system with two axes: The x axis represents
  // an input RGB component or intensity, the y axis an output RGB
  // component or intensity. The ranges of input and output values go 
  // from 0 - 255. 
  // If I take a straight line that intersects (0,0) through (255,255), 
  // I have defined a Null filter that leaves the image unchanged. 

  // Next step: Let's rotate this straight line around the point (125,125).
  // If I increase the slope, output values above (125,125) are increased 
  // by the slope factor, output values below (125, 125) are decreased 
  // likewise. This is how I create the contrast enhancement effect.

  // Now we are having a problem: Increasing the slope means that our
  // straight line intersects the line (x, 255), creating output values 
  // above 255 which are not defined in our RGB model. Since in the RGB world
  // there is nothing whiter than white, I'll set all these values to 255. A 
  // similar thing happens  on the lower left of our imaginary diagram: All
  // values smaller than 0 have to be set 0 - according to the fact that 
  // there is nothing darker than absolute darkness...

  // Ok, now what is that offset value good for? Remember the point 
  // (125,125). At this position, input and output values of our filter 
  // remain unchanged, independent of the slope of our straight line. Now 
  // imagine we are shifting our line parallel to the x axis. Our filter 
  // works quite differently now: We create a cut-off for the higher output 
  // value range (shift right), or the lower ones (shift left).

  INT total = 0;
  register int inc;

  double contrast = m_contrast;
 
  PLASSERT (pBmpSource->GetBitsPerPixel() >= 24);

  pBmpDest->Create (pBmpSource->GetWidth(), 
                    pBmpSource->GetHeight(),
                    pBmpSource->GetBitsPerPixel(),
                    pBmpSource->HasAlpha());

  BYTE ** pSrcLines = pBmpSource->GetLineArray();
  BYTE ** pDstLines = pBmpDest->GetLineArray();

  register int destWidth = pBmpDest->GetWidth();

  if(pBmpSource->GetBitsPerPixel() == 24)
   inc = 3;

  if(pBmpSource->GetBitsPerPixel() == 32)
   inc = sizeof(CPixel32);

  register double red, green, blue;
  register double csupp = contrast * (m_offset - 125.0) + 125.0;

  for (int y = 0; y < pBmpDest->GetHeight(); ++y)
  { // For each line
    register BYTE * pSrcPixel = pSrcLines[y];
    register BYTE * pDstPixel = pDstLines[y];

    for (register int x = 0; x < destWidth; ++x)
    {
       // Formel fr Kontrastberechnung:
       // v = (contrast * (v - 125.0 + m_offset) + 125.0);
       red   = contrast * ((double) (pSrcPixel[RGBA_RED])) + csupp;
       green = contrast * ((double) (pSrcPixel[RGBA_GREEN])) + csupp;
       blue  = contrast * ((double) (pSrcPixel[RGBA_BLUE])) + csupp;

       if(red >= 255.0)
        pDstPixel[RGBA_RED] = (BYTE) 255;
       else if (red < 0.0)
        pDstPixel[RGBA_RED] = (BYTE) 0;
       else
        pDstPixel[RGBA_RED] = (BYTE) red;

       if(green >= 255.0)
        pDstPixel[RGBA_GREEN] = (BYTE) 255;
       else if (green < 0.0)
        pDstPixel[RGBA_GREEN] = (BYTE) 0;
       else
        pDstPixel[RGBA_GREEN] = (BYTE) green;

       if(blue >= 255.0)
        pDstPixel[RGBA_BLUE] = (BYTE) 255;
       else if (blue < 0.0)
        pDstPixel[RGBA_BLUE] = (BYTE) 0;
       else
        pDstPixel[RGBA_BLUE] = (BYTE) blue;

       pSrcPixel += inc;
       pDstPixel += inc;
    }
  }
}

/*
/--------------------------------------------------------------------
|
|      $Log: FilterContrast.cpp,v $
|      Revision 1.3  2001/01/12 23:30:01  uzadow
|      Doc update.
|
|      Revision 1.2  2000/12/18 22:42:53  uzadow
|      Replaced RGBAPIXEL with CPixel32.
|
|      Revision 1.1  2000/11/06 23:20:22  uzadow
|      Added Contrast, Intensity and Lightness filters by
|      Thomas Hirschmann
|
|
\--------------------------------------------------------------------
*/
