/*
/--------------------------------------------------------------------
|
|      $Id: FilterQuantize.cpp,v 1.4 2000/12/18 22:42:53 uzadow Exp $
|
|      Copyright (c) 1996-1998 Ulrich von Zadow
|
|      Based on code:
|
|        Copyright (C) 1998 - 2000 by Punk Productions Electronic Entertainment.
|                                http://cust.nol.at/ppee                                    
|        Content: Quantization/Dithering
|        Programmer: Nikolaus Brennig (virtualnik@nol.at)
|
|      and
|
|        DIBQuant version 1.0
|        Copyright (c) 1993 Edward McCreary.
|        All rights reserved.
|
|        Redistribution and use in source and binary forms are freely permitted
|        provided that the above copyright notice and attibution and date of work
|        and this paragraph are duplicated in all such forms.
|        THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
|        IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
|        WARRANTIES OF MERCHANTIBILILTY AND FITNESS FOR A PARTICULAR PURPOSE.
|
\--------------------------------------------------------------------
*/

#include "stdpch.h"
#include "FilterQuantize.h"
#include "bitmap.h"
#include "anybmp.h"
#include "except.h"

#define COLOR_MAX			32

CFilterQuantize::CFilterQuantize (int DitherPaletteType, int DitherType)
  : m_DitherPaletteType (DitherPaletteType),
    m_DitherType (DitherType)
{
}

CFilterQuantize::~CFilterQuantize()
{

}

void CFilterQuantize::Apply(CBmp * pBmpSource, CBmp * pBmpDest) 
{
  BYTE ** pDstLines = pBmpDest->GetLineArray();

  // Only works for 32 bpp bitmaps.
  PLASSERT (pBmpSource->GetBitsPerPixel() == 32);

  pBmpDest->Create (pBmpSource->GetWidth(), pBmpSource->GetHeight(), 8, FALSE);

  initLUT();

  if (m_DitherPaletteType != DTHPAL_DEFAULT)
    genColorArray(pBmpSource);      

  // Generate the palette.
  switch (m_DitherPaletteType)
  {
    case DTHPAL_MEDIAN:
      genMedianPalette (pBmpSource, pBmpDest);
      break;
    case DTHPAL_POPULARITY:    
      genPopularityPalette(pBmpSource, pBmpDest);                          
      break;
    case DTHPAL_DEFAULT:               
      genDefaultPalette(pBmpDest);
      break;
    default:
      // Incorrect palette type specified in constructor.
      PLASSERT (FALSE);
  }

  ditherDestBmp(pBmpSource, pBmpDest);

  deleteLUT();
}


// Initialize local variables
void CFilterQuantize::initLUT()
{
  ULONG i; 

  m_pQuBoxes = new QUBOX[256];
  if (!m_pQuBoxes) 
    throw (CTextException (ERR_NO_MEMORY, "Out of memory."));
  
  ULONG len = (long)COLOR_MAX*(long)COLOR_MAX*(long)COLOR_MAX;
 
  // buffer to store every color in image
	m_ppHisto = new HISTONODE*[len];
  for (i=0; i<len; i++)
    m_ppHisto[i] = NULL;
}

void CFilterQuantize::deleteLUT()
{
  delete[] m_pQuBoxes;

  int len = (long)COLOR_MAX*(long)COLOR_MAX*(long)COLOR_MAX;
 
  int i;
  if (m_ppHisto)
    for (i=0; i<len; i++)
      if (m_ppHisto[i])
        delete m_ppHisto[i];

  delete[] m_ppHisto;
}


void CFilterQuantize::genMedianPalette (CBmp * pBmpSource, CBmp * pBmpDest)
{
  //--------------------------------------------------------------
  // build new palette with median cut algorithm
  // I didn't write this, if you know who did, please let me know!
  //                                               Ed McCreary
  //--------------------------------------------------------------

  /* force the counts in the corners to be zero */
  addColor (CPixel32 (0,0,0,0), 0L);
  addColor (CPixel32 (31,0,0,0), 0L);
  addColor (CPixel32 (0,31,0,0), 0L);
  addColor (CPixel32 (0,0,31,0), 0L);
  addColor (CPixel32 (31,31,0,0), 0L);
  addColor (CPixel32 (0,31,31,0), 0L);
  addColor (CPixel32 (31,0,31,0), 0L);
  addColor (CPixel32 (31,31,31,0), 0L);

  /* assign the 1st eight boxes to be the corners */
  makeBox (CPixel32 (0,0,0,0), 0, 1L);
  makeBox (CPixel32 (31,0,0,0), 1, 1L);
  makeBox (CPixel32 (0,31,0,0), 2, 1L);
  makeBox (CPixel32 (0,0,31,0), 3, 1L);
  makeBox (CPixel32 (31,31,0,0), 4, 1L);
  makeBox (CPixel32 (0,31,31,0), 5, 1L);
  makeBox (CPixel32 (31,0,31,0), 6, 1L);
  makeBox (CPixel32 (31,31,31,0), 7, 1L);

  /* set up 9th box to hold the rest of the world */
  m_pQuBoxes[8].Corner0 = CPixel32 (0,0,0,0);
  m_pQuBoxes[8].Corner1 = CPixel32 (31,31,31,0);
  squeeze(&m_pQuBoxes[8]);


  /* split the rest of the boxes */
  int i;
  for(i=9; i<256; i++)    
  {
    /* find biggest box */
    int Biggest = 8;
    int j;
    for( j=8; j<i; j++ )
      if(m_pQuBoxes[j].count > m_pQuBoxes[Biggest].count) 
        Biggest = j;
    if (m_pQuBoxes[Biggest].count > 1)
    {
      /* decide which side to split the box along, and split it */
      CPixel32 Corner0 = m_pQuBoxes[Biggest].Corner0; 
      CPixel32 Corner1 = m_pQuBoxes[Biggest].Corner1; 
      int dr = Corner1.GetR () - Corner0.GetR ();
      int dg = Corner1.GetG () - Corner0.GetG ();
      int db = Corner1.GetB () - Corner0.GetB ();
      m_pQuBoxes[i] = m_pQuBoxes[Biggest];      /* copy info over */

      // Determine at which color component to split.
      int ColComp;
      if(dr>=dg && dr>=db) 
        ColComp = RGBA_RED;
      else if(dg>=db) 
        ColComp = RGBA_GREEN;
      else
        ColComp = RGBA_BLUE;
    
      split (&m_pQuBoxes[i],&m_pQuBoxes[Biggest], ColComp);
    }
    else
    { // The biggest box has size 1. This means we've found all the colors.
      // Set rest of the palette to black.
      for (j=i; j<256; j++)
      {
        m_pQuBoxes[j].Corner0 = CPixel32 (0,0,0,0);
        m_pQuBoxes[j].Corner1 = CPixel32 (0,0,0,0);
        m_pQuBoxes[j].Average = CPixel32 (0,0,0,0);
        m_pQuBoxes[j].count = 0;
      }

      // And make the loop exit.
      i = 256;
    }
  } /* end of i loop, we've found 256 boxes. */

  
  /* Generate the actual palette */
  CPixel32 * pClrTab = pBmpDest->GetPalette();

  for( i=0; i<256; i++ )
  {
    // For each color component, find the average of the two corners of the box and
    // scale by the amount of colors in the palette.
    BYTE r = ((long)(m_pQuBoxes[i].Corner0.GetR ()+m_pQuBoxes[i].Corner1.GetR ()))*255/(2*31);
    BYTE g = ((long)(m_pQuBoxes[i].Corner0.GetG ()+m_pQuBoxes[i].Corner1.GetG ()))*255/(2*31);
    BYTE b = ((long)(m_pQuBoxes[i].Corner0.GetB ()+m_pQuBoxes[i].Corner1.GetB ()))*255/(2*31);
    pClrTab[i].Set (r,g,b,0);
  }
}

// Split a box into two at the average color.
// (The actual split is done before the function is called. split() just
// calculates new upper and lower bounds for the two boxes.)
// ColComp is the color component that the split is done on, i.e. one of
// RGBA_RED, RGBA_GREEN, RGBA_BLUE
void CFilterQuantize::split (QUBOX * pBox0, QUBOX * pBox1, int ColComp)
{
  // If we had good accessor functions for the color components, this 
  // would be more readable.
  BYTE * pB0C0 = (BYTE *)&(pBox0->Corner0);  // Box0, Color 0
  BYTE * pB0C1 = (BYTE *)&(pBox0->Corner1);
  BYTE * pB0Ave = (BYTE *)&(pBox0->Average);
  BYTE * pB1C0 = (BYTE *)&(pBox1->Corner0);  
  BYTE * pB1C1 = (BYTE *)&(pBox1->Corner1);
  BYTE * pB1Ave = (BYTE *)&(pBox1->Average);

  if (pB0C1[ColComp]-pB0C0[ColComp] == 2) 
  { /* tight squeeze */
    pB0C1[ColComp] = pB0C0[ColComp];
    pB1C0[ColComp] = pB1C1[ColComp];
  } 
  else 
  { /* figure out where to split */
    BYTE j = pB1Ave[ColComp];
    if (j == pB1C1[ColComp])
      j--;
    pB1C1[ColComp] = j;
    pB0C0[ColComp] = j+1;
  }
  PLASSERT (pB0C0[ColComp] < 32);
  PLASSERT (pB1C0[ColComp] < 32);
  PLASSERT (pB0C1[ColComp] < 32);
  PLASSERT (pB1C1[ColComp] < 32);

  squeeze (pBox0);
  squeeze (pBox1);
  
}

// shrink a boxes extremes to fit tightly
// if a box is 1x1x1 change its count to 1
void CFilterQuantize::squeeze(QUBOX * pBox)
{
  CPixel32 Corner0, Corner1;  

  Corner0 = pBox->Corner0;
  Corner1 = pBox->Corner1;

  pBox->Corner0 = CPixel32 (31,31,31,0);
  pBox->Corner1 = CPixel32 (0,0,0,0);
  pBox->Average = CPixel32 (0,0,0,0);

  int r,g,b;
  int rsum = 0;
  int gsum = 0;
  int bsum = 0;
  ULONG count = 0;
  for (r=Corner0.GetR (); r<=Corner1.GetR (); r++)
  {
    for (g=Corner0.GetG (); g<=Corner1.GetG (); g++)
    {
      for (b=Corner0.GetB (); b<=Corner1.GetB (); b++)
      {
        int index = getColorTableIndex (CPixel32 (r,g,b,0));
        HISTONODE * pNode = m_ppHisto[index];
        if (pNode) 
        {
          if (pNode->count>0L) 
          {
            pBox->Corner0 = CPixel32 (min (r, (int)pBox->Corner0.GetR ()),
                                      min (g, (int)pBox->Corner0.GetG ()),
                                      min (b, (int)pBox->Corner0.GetB ()),
                                      0);
            pBox->Corner1 = CPixel32 (max (r, (int)pBox->Corner1.GetR ()),
                                      max (g, (int)pBox->Corner1.GetG ()),
                                      max (b, (int)pBox->Corner1.GetB ()),
                                      0);
            rsum += r*pNode->count;
            gsum += g*pNode->count;
            bsum += b*pNode->count;

            count += pNode->count;
          }
        }
      }
    }
  }

  /* box is now shrunk */
  if (count)
    pBox->Average = CPixel32 ((BYTE)(rsum / count),
                              (BYTE)(gsum / count),
                              (BYTE)(bsum / count),
                              0);

  pBox->count = min (count, (ULONG)0xFFFFL);

  if (pBox->Corner0 == pBox->Corner1)
  { /* box is min size */
    pBox->count = 1;  /* so it won't get split again */
  }
}

void SwapLong (long* pa, long* pb)
{
  long tmp = *pa;
  *pa = *pb;
  *pb = tmp;
}

void CFilterQuantize::genPopularityPalette (CBmp * pBmpSource, CBmp * pBmpDest)
{
  ULONG PalCount[256];
  memset (PalCount, 0, sizeof(ULONG)*256);
  CPixel32 * pPal = pBmpDest->GetPalette();
  
  /* force corners of rgb color cube out of the running */
  addColor (CPixel32 (0,0,0,0), 0L);
  addColor (CPixel32 (31,0,0,0), 0L);
  addColor (CPixel32 (0,31,0,0), 0L);
  addColor (CPixel32 (0,0,31,0), 0L);
  addColor (CPixel32 (31,31,0,0), 0L);
  addColor (CPixel32 (0,31,31,0), 0L);
  addColor (CPixel32 (31,0,31,0), 0L);
  addColor (CPixel32 (31,31,31,0), 0L);

  /* force feed the corners into the palette */
  pPal[0] = CPixel32 (0,0,0,0);
  pPal[1] = CPixel32 (31,0,0,0);
  pPal[2] = CPixel32 (0,31,0,0);
  pPal[3] = CPixel32 (0,0,31,0);
  pPal[4] = CPixel32 (31,31,0,0);
  pPal[5] = CPixel32 (0,31,31,0);
  pPal[6] = CPixel32 (31,0,31,0);
  pPal[7] = CPixel32 (31,31,31,0);

  int r,g,b;
  int IndexCache = -1;
  HISTONODE * pNode;
  HISTONODE * pNodeCache;
  for (r=0; r<31; r++)
  {
    for (g=0; g<31; g++) 
    {
      for (b=0; b<31; b++) 
      {             
        int index = getColorTableIndex (CPixel32 (r,g,b,0));
        if (index == IndexCache)
          pNode = pNodeCache;
        else
        {
          pNode = m_ppHisto[index];
          pNodeCache = pNode;
          IndexCache = index;
        }

        if (pNode) 
        {
          if(pNode->count > PalCount[255]) 
          {
            PalCount[255] = pNode->count;
            pPal[255].Set (r,g,b,0);

            int i = 255;        /* bubble up */
            while(PalCount[i] > PalCount[i-1] && i>8) 
            {
              SwapLong ((long *)&PalCount[i], (long*)&PalCount[i-1]);
              SwapLong ((long *)&pPal[i], (long *)&pPal[i-1]);
              i--;
            }
          }
        }                
      }       /* end of current chain */
    }       /* end of r loop */            
  }   
      
  int i;
  for (i=0; i<256; i++)
  {
    pPal[i].SetR ((int(pPal[i].GetR ())*255)/31);
    pPal[i].SetG ((int(pPal[i].GetG ())*255)/31);
    pPal[i].SetB ((int(pPal[i].GetB ())*255)/31);
  }
}

void CFilterQuantize::genColorArray (CBmp * pBmpSource)
{
  int x,y;
  int h = pBmpSource->GetHeight();
  int w = pBmpSource->GetWidth();
  BYTE ** pSrcLines = pBmpSource->GetLineArray();
  for (y=0; y<h; y++)
  {
    CPixel32 * pSrcPixel;
    pSrcPixel = (CPixel32 *)pSrcLines[y];
    for (x=0; x<w; x++)
    {
      addColor(*pSrcPixel, 1L);
      pSrcPixel ++;
    }  
  }

}

BYTE DefaultPalette[] = 
{
	0,0,0,
	191,0,0,
	0,191,0,
	191,191,0,
	0,0,191,
	191,0,191,
	0,191,191,
	192,192,192,
	192,220,192,
	166,202,240,
	128,0,0,
	0,128,0,
	128,128,0,
	0,0,128,
	128,0,128,
	0,128,128,
	0,81,0,
	81,81,0,
	133,81,0,
	177,81,0,
	217,81,0,
	255,81,0,
	0,133,0,
	81,133,0,
	133,133,0,
	177,133,0,
	217,133,0,
	255,133,0,
	0,177,0,
	81,177,0,
	133,177,0,
	177,177,0,
	217,177,0,
	255,177,0,
	0,217,0,
	81,217,0,
	133,217,0,
	177,217,0,
	217,217,0,
	255,217,0,
	0,255,0,
	81,255,0,
	133,255,0,
	177,255,0,
	217,255,0,
	255,255,0,
	0,0,81,
	81,0,81,
	133,0,81,
	177,0,81,
	217,0,81,
	255,0,81,
	0,81,81,
	81,81,81,
	133,81,81,
	177,81,81,
	217,81,81,
	255,81,81,
	0,133,81,
	81,133,81,
	133,133,81,
	177,133,81,
	217,133,81,
	55,133,81,
	0,177,81,
	81,177,81,
	133,177,81,
	177,177,81,
	217,177,81,
	255,177,81,
	0,217,81,
	81,217,81,
	133,217,81,
	177,217,81,
	217,217,81,
	255,217,81,
	0,255,81,
	81,255,81,
	133,255,81,
	177,255,81,
	217,255,81,
	255,255,81,
	0,0,133,
	81,0,133,
	133,0,133,
	177,0,133,
	217,0,133,
	255,0,133,
	0,81,133,
	81,81,133,
	133,81,133,
	177,81,133,
	217,81,133,
	255,81,133,
	0,133,133,
	81,133,133,
	133,133,133,
	177,133,133,
	217,133,133,
	255,133,133,
	0,177,133,
	81,177,133,
	133,177,133,
	177,177,133,
	217,177,133,
	255,177,133,
	0,217,133,
	81,217,133,
	133,217,133,
	177,217,133,
	217,217,133,
	255,217,133,
	0,255,133,
	81,255,133,
	133,255,133,
	177,255,133,
	217,255,133,
	255,255,133,
	0,0,177,
	81,0,177,
	133,0,177,
	177,0,177,
	217,0,177,
	255,0,177,
	0,81,177,
	81,81,177,
	133,81,177,
	177,81,177,
	217,81,177,
	255,81,177,
	0,133,177,
	81,133,177,
	133,133,177,
	177,133,177,
	217,133,177,
	255,133,177,
	0,177,177,
	81,177,177,
	133,177,177,
	177,177,177,
	217,177,177,
	255,177,177,
	0,217,177,
	81,217,177,
	133,217,177,
	177,217,177,
	217,217,177,
	255,217,177,
	0,255,177,
	81,255,177,
	133,255,177,
	177,255,177,
	217,255,177,
	255,255,177,
	0,0,217,
	81,0,217,
	133,0,217,
	177,0,217,
	217,0,217,
	255,0,217,
	0,81,217,
	81,81,217,
	133,81,217,
	177,81,217,
	217,81,217,
	255,81,217,
	0,133,217,
	81,133,217,
	133,133,217,
	177,133,217,
	217,133,217,
	255,133,217,
	0,177,217,
	81,177,217,
	133,177,217,
	177,177,217,
	217,177,217,
	255,177,217,
	0,217,217,
	81,217,217,
	133,217,217,
	177,217,217,
	217,217,217,
	255,217,217,
	0,255,217,
	81,255,217,
	133,255,217,
	177,255,217,
	217,255,217,
	255,255,217,
	0,0,255,
	81,0,255,
	133,0,255,
	177,0,255,
	217,0,255,
	255,0,255,
	0,81,255,
	81,81,255,
	133,81,255,
	177,81,255,
	217,81,255,
	255,81,255,
	0,133,255,
	81,133,255,
	133,133,255,
	177,133,255,
	217,133,255,
	255,133,255,
	0,177,255,
	81,177,255,
	133,177,255,
	177,177,255,
	217,177,255,
	255,177,255,
	0,217,255,
	81,217,255,
	133,217,255,
	177,217,255,
	217,217,255,
	255,217,255,
	0,255,255,
	81,255,255,
	133,255,255,
	177,255,255,
	217,255,255,
	196,196,196,
	199,199,199,
	202,202,202,
	205,205,205,
	208,208,208,
	211,211,211,
	214,214,214,
	217,217,217,
	220,220,220,
	223,223,223,
	226,226,226,
	229,229,229,
	232,232,232,
	235,235,235,
	238,238,238,
	241,241,241,
	244,244,244,
	247,247,247,
	250,250,250,
	253,253,253,
	155,207,207,
	255,251,240,
	160,160,164,
	128,128,128,
	255,0,0,
	0,255,0,
	255,255,0,
	0,0,255,
	255,0,255,
	0,255,255,
	255,255,255
};

void CFilterQuantize::genDefaultPalette (CBmp * pBmpDest)
{
  CPixel32 * pPal = pBmpDest->GetPalette();
  int i;
  for( i=0; i<256; i++ )
    memcpy (&(pPal[i]), &(DefaultPalette[i*3]), 3);
}

// Adds a RGB Color into the color histogram.
void CFilterQuantize::addColor(CPixel32 col, ULONG count)
{
  HISTONODE * pNode;
    
  ULONG index = getShiftedColorTableIndex(col);
  count = min (count, (ULONG)0xFFFFL);
  
  pNode = m_ppHisto[index];
  if(!pNode)  // new color
  {
    pNode = m_ppHisto[index] = new HISTONODE;
    pNode->index = -1;
    pNode->count = count;
  }
  else
  {    
    ULONG NewCount;
    NewCount = pNode->count;
    NewCount += count;
    pNode->count = min (NewCount, (ULONG)0xFFFFL);
  }
}

// make a 1x1x1 box at index with color rgb count c
void CFilterQuantize::makeBox(CPixel32 col, int i, ULONG c)
{
  m_pQuBoxes[i].Corner0 = col;
  m_pQuBoxes[i].Corner1 = col;
  m_pQuBoxes[i].count = c;
}

int CFilterQuantize::getColorTableIndex (CPixel32 col)
{
  DWORD b = col.GetB ();
  DWORD g = col.GetG ();
  DWORD r = col.GetR ();
  return (b | g << 5 | r << 10);
}

int CFilterQuantize::getShiftedColorTableIndex (CPixel32 col)
{
  DWORD b = col.GetB ()>>3;
  DWORD g = col.GetG ()>>3;
  DWORD r = col.GetR ()>>3;
  return (b | g << 5 | r << 10);
}

void CFilterQuantize::ditherDestBmp(CBmp * pBmpSource, CBmp * pBmpDest)
{

  BYTE ** ppSrcLines = pBmpSource->GetLineArray();
  BYTE ** ppDestLines = pBmpDest->GetLineArray();

  int w = pBmpSource->GetWidth();
  int h = pBmpSource->GetHeight();

  // FS error accumulators.
  double * pOddErrors;
  double * pEvenErrors;
  double * pCurErrors;
  double * pNextErrors;

  // FS: current pixel as float values
  double r,g,b;

  if (m_DitherType == DTH_FS)
  {
    // Get memory for errors
    pOddErrors = new double[(w + 2) * 6];
    pEvenErrors = &pOddErrors[(w + 2) * 3];
    memset(pOddErrors, 0, (w + 2) * 6 * sizeof(double));
  }
  
  // Loop row by row, col by col, and of course dither...
  int y;
  int i;
  int SrcInc, DestInc;
  for (y=0; y<h; y++)
  {
    BYTE * pSrcLine = ppSrcLines[y];
    BYTE * pDestLine = ppDestLines[y];
    if (m_DitherType == DTH_FS)
    { // Floyd-Steinberg dithers in different directions in alternate rows.
      if (y&1) // odd row
      {
        // Start at last pixel in row.
        pSrcLine += w*4-4;
        pDestLine += w-1;
        SrcInc = -4;
        DestInc = -1;

        pCurErrors = pOddErrors + 3;
        pNextErrors = pEvenErrors + (w * 3);
      }
      else // even row
      {
        SrcInc = 4;
        DestInc = 1;

        pCurErrors = pEvenErrors + 3;
        pNextErrors = pOddErrors + (w * 3);
      }

      // Initialize the error constants for the first part of the row
      pNextErrors[0] = 0;
      pNextErrors[1] = 0;
      pNextErrors[2] = 0;
    }
    else
    {
      SrcInc = 4;
      DestInc = 1;
    }
    
    // go through the scanline...
    i = w;
    while(i--)
    {    
      // Get RGB Values...
      CPixel32 pixel = *(CPixel32 *)pSrcLine;
      pSrcLine += SrcInc;
     
      // Now dither...
      switch (m_DitherType)
      {
/*        case DTH_JITTER:
          jitterPixel( i, y, &pixel);
          break;
*/
        case DTH_ORDERED:
          ditherPixelOrdered (i, y, &pixel);
          break;
        case DTH_FS:
          r = pixel.GetR ();
          g = pixel.GetG ();
          b = pixel.GetB ();
          ditherPixelFS(&r, &g, &b, pCurErrors);
          pixel.Set ((BYTE)r, (BYTE)g, (BYTE)b, 0);
          break;
      }

      // Now get the best possible paletteindex matching the RGB Vals.
      BYTE index = getNeighbor (pixel, pBmpDest->GetPalette());

      // FS dithering now calculates the error we've introduced by 
      // approximating.
      if (m_DitherType == DTH_FS)
      {
        CPixel32 * pPal = pBmpDest->GetPalette();
        double PalR = (double)(pPal[index].GetR ());
        double PalG = (double)(pPal[index].GetG ());
        double PalB = (double)(pPal[index].GetB ());

        double error;
        
        // Red component
        error = (r - PalR) / 16.0;
        pNextErrors[-3] = error; 
        pNextErrors[+3] += error*3;
        pNextErrors[+0] += error*5;
        pCurErrors[+3] += error*7;

        // Green component
        error = (g - PalG) / 16.0;
        pNextErrors[-2]  = error; 
        pNextErrors[+4] += error*3;
        pNextErrors[+1] += error*5;
        pCurErrors[+4] += error*7;

        // Blue component
        error = (b - PalB) / 16.0;
        pNextErrors[-1]  = error; 
        pNextErrors[+5] += error*3;
        pNextErrors[+2] += error*5;
        pCurErrors[+5] += error*7;

        pNextErrors -= 3;
        pCurErrors += 3;
      }

      // Set the pixel in the destination bitmap
      *pDestLine = (BYTE)index;
      pDestLine += DestInc;
    }  

  }

  // clean up.
  if (m_DitherType == DTH_FS ) 
    delete [] pOddErrors;
   
}


void CFilterQuantize::jitterPixel (int i, int y, CPixel32 * pPixel)
{
}

void CFilterQuantize::ditherPixelOrdered (int x, int y, CPixel32 * pPixel)
{
  BYTE * pPx = (BYTE *)pPixel;
  ditherCompOrdered (x, y, &pPx[RGBA_RED]);
  ditherCompOrdered (x, y, &pPx[RGBA_GREEN]);
  ditherCompOrdered (x, y, &pPx[RGBA_BLUE]);

}

void CFilterQuantize::ditherCompOrdered (int x, int y, BYTE * pComp)
{
  const INT DitherLUT[4][4] = { -7, 1, -5, 3, 5, -3, 7, -1, -4, 4, -6, 2, 8, 0, 6, -2 };

  int Comp = *pComp;
  Comp += DitherLUT [x&3][y&3]<<1;
  Comp = clip (Comp);
  *pComp = (BYTE)Comp;
}

void CFilterQuantize::ditherPixelFS (double * pR, double * pG, double * pB, double * pCurErrors)
{
  ditherCompFS (pR, pCurErrors[0]); 
  ditherCompFS (pG, pCurErrors[1]); 
  ditherCompFS (pB, pCurErrors[2]); 
}

void CFilterQuantize::ditherCompFS (double * pComp, double Error)
{
  *pComp += Error;
  if (*pComp > 255) *pComp = 255;
  if (*pComp < 0) *pComp = 0;
}

// Returns nearest color to the color passed in that's present
// in the palette as a palette index.
BYTE CFilterQuantize::getNeighbor (CPixel32 Color, CPixel32 * pPal)
{
  HISTONODE * pNode;
  

  int TableIndex = getShiftedColorTableIndex (Color);

  pNode = m_ppHisto[TableIndex];

  if (!pNode)
  {
    m_ppHisto[TableIndex] = new HISTONODE;
    pNode = m_ppHisto[TableIndex];
    pNode->count = 0;
    pNode->index = -1;
  }

  int PalIndex = pNode->index;
  if( PalIndex == -1)
  {   
    pNode->index = 0;

    CPixel32 BestColor = pPal[0];
    int MinDist = colorDist (BestColor, Color);

    int c;
    for (c = 1; c<256; c++)
    {
      int Dist = colorDist (pPal[c], Color);
      if (Dist < MinDist)
      {
        pNode->index = c; 
        MinDist = Dist;
      }
    }
    PalIndex = pNode->index;
  }
  return PalIndex;
}

int CFilterQuantize::colorDist (CPixel32 c0, CPixel32 c1)
{
  int dr = c0.GetR ()-c1.GetR ();
  int dg = c0.GetG ()-c1.GetG ();
  int db = c0.GetB ()-c1.GetB ();
  return dr*dr+dg*dg+db*db;
}

/*
/--------------------------------------------------------------------
|
|      $Log: FilterQuantize.cpp,v $
|      Revision 1.4  2000/12/18 22:42:53  uzadow
|      Replaced RGBAPIXEL with CPixel32.
|
|      Revision 1.3  2000/12/09 12:16:26  uzadow
|      Fixed several memory leaks.
|
|      Revision 1.2  2000/10/23 17:45:03  jmbuena
|      Linux compatibility changes
|
|      Revision 1.1  2000/09/26 12:14:49  Administrator
|      Refactored quantization.
|
|
|
\--------------------------------------------------------------------
*/
