/*
/--------------------------------------------------------------------
|
|      $Id: bitmap.cpp,v 1.27 2000/12/18 22:42:52 uzadow Exp $
|      Device independent bitmap class
|
|      Manipulates uncompressed device independent bitmaps
|      of all color depths.
|
|      Copyright (c) 1996-1998 Ulrich von Zadow
|
\--------------------------------------------------------------------
*/

#include "stdpch.h"
#include "bitmap.h"
#include "except.h"
#include "FilterResizeBilinear.h"
#include "FilterResizeBox.h"
#include "FilterResizeGaussian.h"
#include "FilterResizeHamming.h"
#include "FilterCrop.h"
#include "FilterGrayscale.h"
#include "FilterThreshold.h"
#include "FilterGetAlpha.h"
#include "FilterRotate.h"
#include "FilterQuantize.h"
#include "VideoInvertFilter.h"

CBmp::CBmp ()
// Creates an empty bitmap.
// Derived classes have to create a small bitmap here so the
// class can assume that a valid bitmap is available all the
// time.
{
  m_LockCount = 0;
  // Michael Salzlechner - March 02 2000
  //
  // defaults for 8Bit color quantization
  m_DitherPaletteType = DTHPAL_DEFAULT;
  m_DitherType = DTH_NONE;
}


CBmp::~CBmp
    ()
{
  PLASSERT (m_LockCount == 0);
}

bool const CBmp::operator ==
    ( CBmp const &Other
    )
{
  if (GetWidth() != Other.GetWidth() ||
      GetHeight() != Other.GetHeight() ||
      HasAlpha() != Other.HasAlpha() ||
      GetBitsPerPixel() != Other.GetBitsPerPixel())
    return false;

  BYTE ** ppLines1 = GetLineArray();
  BYTE ** ppLines2 = Other.GetLineArray();
  int y,x;
  for (y=0; y<GetHeight(); y++)
    for (x=0; x<GetWidth(); x++)
      switch (GetBitsPerPixel())
      {
        case 8:
          if (ppLines1[y][x] != ppLines2[y][x])
            return false;
          break;
        case 32:
          if (ppLines1[y][x*4+RGBA_RED] != ppLines2[y][x*4+RGBA_RED] ||
              ppLines1[y][x*4+RGBA_GREEN] != ppLines2[y][x*4+RGBA_GREEN] ||
              ppLines1[y][x*4+RGBA_BLUE] != ppLines2[y][x*4+RGBA_BLUE])
            return false;
          if (HasAlpha() &&
              ppLines1[y][x*4+3] != ppLines2[y][x*4]+3)
            return false;
          break;
        default:
          // Bitmap must be either 8 or 32 bpp for this routine to work.
          PLASSERT (false);
      }

  // Test if the palettes are the same for paletted bitmaps.
  if (GetBitsPerPixel() == 8)
  {
    int i;
    CPixel32 * pPal1 = GetPalette();
    CPixel32 * pPal2 = Other.GetPalette();
    for (i=0; i<255; i++)
    {
      if (pPal1[i] != pPal2[i])
        return false;
    }
  }

  return true;
}

void CBmp::CreateCopy (const CBmp & rSrcBmp, int BPPWanted)
// Creates a copy of rSrcBmp, converting color depth if nessesary.
// Supports 1, 8 and 32 BPP. Alpha channel information is preserved.
{
  PLASSERT_VALID (&rSrcBmp);

  PLASSERT (BPPWanted == 32 || BPPWanted == 8 ||
          BPPWanted == 1 || BPPWanted == 0);
  int BPPSrc = rSrcBmp.GetBitsPerPixel();
  PLASSERT (BPPSrc == 32 || BPPSrc == 8 || BPPSrc == 1);

  if (BPPWanted == BPPSrc || BPPWanted == 0)
  {
    if (&rSrcBmp != this)
    {
      // Create empty bitmap.
      freeMembers ();
      internalCopy (rSrcBmp);
    }
  }
  else
  {
    // Can't copy to self while changing bit depth.
    PLASSERT (&rSrcBmp != this);

    BOOL bDestAlpha = rSrcBmp.HasAlpha() && BPPWanted == 32;
    Create (rSrcBmp.GetWidth(), rSrcBmp.GetHeight(),
            BPPWanted, bDestAlpha);

    switch (BPPWanted)
    {
      case 32:
        create32BPPCopy (rSrcBmp);
        break;

      case 8:
        create8BPPCopy (rSrcBmp);
        break;

      case 1:
        create1BPPCopy (rSrcBmp);
        break;
      default:
        PLASSERT(FALSE);
    }
    PLASSERT_VALID (this);
  }
}

void CBmp::CreateGrayscaleCopy (CBmp & rSrcBmp)
{
  PLASSERT_VALID (this);
  PLASSERT_VALID (&rSrcBmp);
  CFilterGrayscale Filter;
  Filter.Apply (&rSrcBmp, this);
}

void CBmp::CreateRotatedCopy (CBmp & rSrcBmp, double angle, CPixel32 color)
{
  PLASSERT_VALID (this);
  PLASSERT_VALID (&rSrcBmp);
  CFilterRotate Filter (angle, color);
  Filter.Apply (&rSrcBmp, this);
}


CPLPoint CBmp::CreateRotatedCopyPoint (CBmp & rSrcBmp, double angle, CPixel32 color,
                                       CPLPoint Point, CPLPoint Pos)
{
 // Point -> Punkt um den gedreht werden soll
 // Pos   -> Aktuelle linke obere Ecke des Source Bildes

 double DAbsolutX,DAbsolutY,DAbsDestMiddleX,DAbsDestMiddleY;

 CreateRotatedCopy(rSrcBmp, angle, color);

 // Mittelpunkte des Source Bildes
 CPLPoint SrcMiddle  = CPLPoint (Pos.x + rSrcBmp.GetWidth() / 2, Pos.y + rSrcBmp.GetHeight() / 2);

 // Absolute Werte zwischen Drehpunkt und Mittelpunkt des Source Bildes
 CPLPoint Absolut = CPLPoint (Point.x - SrcMiddle.x, Point.y - SrcMiddle.y);

   // Neuen Punkt Absolut berechnen
 DAbsolutX = (double)Absolut.x ;
 DAbsolutY = (double)Absolut.y ;
 DAbsDestMiddleX = DAbsolutX * cos(angle * -1) - DAbsolutY * sin(angle * -1);
 DAbsDestMiddleY = DAbsolutX * sin(angle * -1) + DAbsolutY * cos(angle * -1);

    CPLPoint AbsDestMiddle =  CPLPoint ( (int)DAbsDestMiddleX , (int)DAbsDestMiddleY );
 AbsDestMiddle.x *= -1;
 AbsDestMiddle.y *= -1;

   // Neuen Punkt berechnen
 CPLPoint DestMiddle = CPLPoint (Point.x - AbsDestMiddle.x, Point.y - AbsDestMiddle.y);

  // Rckgabe in Rotated
 return(CPLPoint (DestMiddle.x - GetWidth() / 2, DestMiddle.y - GetHeight() / 2));
};

void CBmp::CreateFromAlphaChannel (CBmp & rSrcBmp)
{
  PLASSERT_VALID (this);
  PLASSERT_VALID (&rSrcBmp);

  CFilterGetAlpha Filter;
  Filter.Apply (&rSrcBmp, this);
}


void CBmp::CreateResizedBilinear (CBmp& rSrcBmp, int NewXSize, int NewYSize)
{
  PLASSERT_VALID (this);
  PLASSERT_VALID (&rSrcBmp);
  CFilterResizeBilinear Filter (NewXSize, NewYSize);
  Filter.Apply (&rSrcBmp, this);
}

void CBmp::CreateResizedBox (CBmp& rSrcBmp, int NewXSize, int NewYSize, double NewRadius)
{
  PLASSERT_VALID (this);
  PLASSERT_VALID (&rSrcBmp);
  CFilterResizeBox Filter (NewXSize, NewYSize, NewRadius);
  Filter.Apply (&rSrcBmp, this);
}

void CBmp::CreateResizedGaussian (CBmp& rSrcBmp, int NewXSize, int NewYSize, double NewRadius)
{
  PLASSERT_VALID (this);
  PLASSERT_VALID (&rSrcBmp);
  CFilterResizeGaussian Filter (NewXSize, NewYSize, NewRadius);
  Filter.Apply (&rSrcBmp, this);
}

void CBmp::CreateResizedHamming (CBmp& rSrcBmp, int NewXSize, int NewYSize, double NewRadius)
{
  PLASSERT_VALID (this);
  PLASSERT_VALID (&rSrcBmp);
  CFilterResizeHamming Filter (NewXSize, NewYSize, NewRadius);
  Filter.Apply (&rSrcBmp, this);
}

void CBmp::CreateCropped (CBmp& rSrcBmp, int XMin, int XMax, int YMin, int YMax)
{
  PLASSERT_VALID (this);
  PLASSERT_VALID (&rSrcBmp);
  CFilterCrop Filter (XMin, XMax, YMin, YMax);
  Filter.Apply (&rSrcBmp, this);
}

void CBmp::SetQuantizationMode (int DitherType, int DitherPaletteType)
{
  PLASSERT_VALID (this);
  m_DitherType = DitherType;
  m_DitherPaletteType = DitherPaletteType;
}

#ifdef _DEBUG
void CBmp::AssertValid () const
{
  if (m_Width != 0 || m_Height != 0)
  {
    // Color table only if 8 bpp.
    PLASSERT ((m_bpp > 8) == (m_pClrTab == NULL));
    PLASSERT (m_Height >= 0);
    PLASSERT (m_Width >= 0);

    // Alpha channel only if 32 bpp
    PLASSERT ((m_bpp == 32) || !m_bAlphaChannel);
  }
  PLASSERT (m_LockCount >= 0);
}
#endif


/////////////////////////////////////////////////////////////////////
// CBmp creation

void CBmp::Create (LONG Width, LONG Height, WORD BitsPerPixel, BOOL bAlphaChannel)
// Create a new empty bitmap. Bits are uninitialized.
{
  PLASSERT_VALID (this);

  freeMembers ();
  internalCreate (Width, Height, BitsPerPixel, bAlphaChannel);

  PLASSERT_VALID (this);
}


/////////////////////////////////////////////////////////////////////
// CBmp manipulation

void CBmp::SetGrayPalette ()
// Fills the color table with a grayscale palette.
{
  PLASSERT (m_pClrTab); // Bitmap must contain a palette!

  int i;
  int NumColors = GetNumColors();
  double ColFactor = 255/(NumColors-1);

  for (i=0; i<NumColors; i++)
    SetPaletteEntry (i, int(i*ColFactor), int(i*ColFactor), int(i*ColFactor), 0xFF);
}

void CBmp::SetPalette (CPixel32 * pPal)
{
  PLASSERT (m_pClrTab); // Bitmap must contain a palette!

  // bdelmee code change
  memcpy (m_pClrTab, pPal, GetNumColors() * sizeof(CPixel32));
}


void CBmp::SetAlphaChannel (CBmp * pAlphaBmp)
// Replaces the alpha channel with a new one.
{
  BYTE * pLine;
  BYTE * pAlphaLine;
  int x,y;
  BYTE ** pLineArray;
  BYTE ** pAlphaLineArray;

  PLASSERT_VALID (this);
  // Current bitmap must be 32 bpp.
  PLASSERT (GetBitsPerPixel() == 32);

  PLASSERT_VALID (pAlphaBmp);
  // Alpha channel must be 8 bpp.
  PLASSERT (pAlphaBmp->GetBitsPerPixel() == 8);

  // The two bitmaps must have the same dimensions
  PLASSERT (pAlphaBmp->GetWidth() == GetWidth());
  PLASSERT (pAlphaBmp->GetHeight() == GetHeight());

  Lock(false, true);
  pAlphaBmp->Lock(true, false) ;
  pLineArray = GetLineArray();
  pAlphaLineArray = pAlphaBmp->GetLineArray();

  for (y=0; y < GetHeight(); y++)
  {
    pLine = pLineArray[y];
    pAlphaLine = pAlphaLineArray[y];
    for (x=0; x < GetWidth(); x++)
    {
      pLine[x*4+RGBA_ALPHA] = pAlphaLine[x];
    }
  }
  pAlphaBmp->Unlock();
  Unlock();

  m_bAlphaChannel = TRUE;

  PLASSERT_VALID (this);
}

void CBmp::ResizeBilinear (int NewXSize, int NewYSize)
{
  PLASSERT_VALID (this);
  CFilterResizeBilinear Filter (NewXSize, NewYSize);
  Filter.ApplyInPlace (this);
}

void CBmp::ResizeBox (int NewXSize, int NewYSize, double NewRadius)
{
  PLASSERT_VALID (this);
  CFilterResizeBox Filter (NewXSize, NewYSize, NewRadius);
  Filter.ApplyInPlace (this);
}

void CBmp::ResizeGaussian (int NewXSize, int NewYSize, double NewRadius)
{
  PLASSERT_VALID (this);
  CFilterResizeGaussian Filter (NewXSize, NewYSize, NewRadius);
  Filter.ApplyInPlace (this);
}

void CBmp::ResizeHamming (int NewXSize, int NewYSize, double NewRadius)
{
  PLASSERT_VALID (this);
  CFilterResizeHamming Filter (NewXSize, NewYSize, NewRadius);
  Filter.ApplyInPlace (this);
}

void CBmp::Crop (int XMin, int XMax, int YMin, int YMax)
{
  PLASSERT_VALID (this);
  CFilterCrop Filter (XMin, XMax, YMin, YMax);
  Filter.ApplyInPlace (this);
}

void CBmp::MakeGrayscale ()
{
  PLASSERT_VALID (this);
  CFilterGrayscale Filter;
  Filter.ApplyInPlace (this);
}

void CBmp::Threshold(int threshold_min, int threshold_max, int channel)
{
  PLASSERT_VALID (this);
  CFilterThreshold Filter (threshold_min, threshold_max, channel);
  Filter.ApplyInPlace (this);

}

void CBmp::Rotate (double angle, CPixel32 color)
{
  PLASSERT_VALID (this);
  CFilterRotate Filter (angle, color);
  Filter.ApplyInPlace (this);
}

void CBmp::Invert ()
{
  PLASSERT_VALID (this);
  CVideoInvertFilter Filter;
  Filter.ApplyInPlace (this);
}

CPixel32 CBmp::GetPixel (int x, int y)
{
  PLASSERT (GetBitsPerPixel() == 32);
  return *((CPixel32*)GetLineArray()[y]+x);
}

void CBmp::SetPixel (int x, int y, CPixel32 pixel)
{
  PLASSERT (GetBitsPerPixel() == 32);
  *((CPixel32*)GetLineArray()[y]+x) = pixel;
}

void CBmp::Lock (bool bReadable, bool bWriteable)
{
  PLASSERT (m_LockCount >= 0);
  PLASSERT (bReadable || bWriteable);
  m_LockCount++;
}

void CBmp::Unlock
    ()
{
  m_LockCount--;
  PLASSERT (m_LockCount >= 0);
}

BYTE CBmp::FindNearestColor (CPixel32 col)
{
  CPixel32 * pPalette = GetPalette();
  PLASSERT (pPalette);

  int d1;
  int dMin = 100000;
  int i;
  int index;
  for (i = 0; i<GetNumColors(); i++)
  {
    d1 = col.BoxDist (pPalette[i]);
    if (d1 < dMin)
    {
      dMin = d1;
      index = i;
    }
  }
  return index;
}

/////////////////////////////////////////////////////////////////////
// Local functions


void CBmp::initLocals (LONG Width, LONG Height, WORD BitsPerPixel, BOOL bAlphaChannel)
{
  m_Width = Width;
  m_Height = Height;
  m_bpp = BitsPerPixel;
  m_bAlphaChannel = bAlphaChannel;

  // Initialize pointers to lines.
  initLineArray ();

  if (BitsPerPixel < 16)
  { // Color table exists
    SetGrayPalette ();
  }

  PLASSERT_VALID (this);
}

void CBmp::internalCopy (const CBmp & rSrcBmp)
{
  PLASSERT_VALID(&rSrcBmp);

  // Create empty bitmap.
  internalCreate (rSrcBmp.GetWidth(), rSrcBmp.GetHeight(), rSrcBmp.GetBitsPerPixel(),
                  rSrcBmp.HasAlpha());
  // We'll change rSrcBmp back to the original state before returning,
  // so the const_cast is ok. I think.
  const_cast<CBmp &>(rSrcBmp).Lock(true, false);
  Lock(false, true);
  BYTE ** pSrcLines = rSrcBmp.GetLineArray();
  BYTE ** pDstLines = GetLineArray();
  // Minimum possible line length.
  int LineLen = GetBytesPerLine();

  for (int y = 0; y<GetHeight(); y++)
    memcpy (pDstLines[y], pSrcLines[y], LineLen);

  Unlock();
  const_cast<CBmp &>(rSrcBmp).Unlock();

  if (GetBitsPerPixel() <= 8)
    SetPalette (rSrcBmp.GetPalette());

  PLASSERT_VALID (this);
}

void CBmp::create32BPPCopy (const CBmp & rSrcBmp)
{
  int BPPSrc = rSrcBmp.GetBitsPerPixel();
  Lock(false, true);
  // We'll change rSrcBmp back to the original state before returning,
  // so the const_cast is ok. I think.
  const_cast<CBmp &>(rSrcBmp).Lock(true, false);
  BYTE ** pSrcLines = rSrcBmp.GetLineArray();
  BYTE ** pDstLines = GetLineArray();
  int SrcLineLen = GetWidth();

  if (BPPSrc == 8) // Conversion 8->32 BPP
  {
    CPixel32 *pPal = rSrcBmp.GetPalette();

    for (int y = 0; y<GetHeight(); ++y)
    { // For each line
      BYTE * pSrcPixel = pSrcLines[y];
      CPixel32 * pDstPixel = (CPixel32 *)pDstLines[y];

      for (int x = 0; x < SrcLineLen; ++x)
      { // For each pixel
        *pDstPixel = pPal[*pSrcPixel];
        ++pSrcPixel;
        ++pDstPixel;
      }
    }
  }
  else // 1 -> 32
  {
    CPixel32 * pPal = rSrcBmp.GetPalette();
    CPixel32 blackDot, whiteDot;
    // if bi-tonal "palette" exists, use it...
    if (pPal)
    {
      whiteDot = pPal[0];
      blackDot = pPal[1];
    }
    else
    {
      whiteDot.Set (255,255,255,255);
      blackDot.Set (0,0,0,255);
    }

    // assume msb is leftmost
    for (int y = 0; y<GetHeight(); ++y)
    { // For each line
      BYTE * pSrcPixel = pSrcLines[y];
      CPixel32 * pDstPixel = (CPixel32 *)pDstLines[y];

      for (int x = 0; x < SrcLineLen; ++x)
      { // For each pixel
        if (pSrcPixel[x / 8] & (128 >> (x & 7)))  // black pixel
          *pDstPixel = blackDot;
        else
          *pDstPixel = whiteDot;
        pDstPixel++;
      }
    }
  }
  const_cast<CBmp &>(rSrcBmp).Unlock();
  Unlock();
}

void CBmp::create8BPPCopy (const CBmp & rSrcBmp)
{
  int BPPSrc = rSrcBmp.GetBitsPerPixel();
  Lock(false, true);
  const_cast<CBmp &>(rSrcBmp).Lock(true, false);

  if (BPPSrc == 32) // Conversion 32->8 BPP
  {
    CFilterQuantize Filter (m_DitherPaletteType, m_DitherType);
    Filter.Apply (&(const_cast<CBmp &>(rSrcBmp)), this);
  }
  else // 1 -> 8
  {
    CPixel32 *pPal = rSrcBmp.GetPalette();
    // if bi-tonal "palette" exists, use it...
    if (pPal)
    {
      BYTE *pWhite = (BYTE *) pPal;
      BYTE *pBlack = (BYTE *) (pPal+1);
      SetPaletteEntry(0,
                      pWhite[RGBA_RED],pWhite[RGBA_GREEN],pWhite[RGBA_BLUE],
                      255);
      SetPaletteEntry(1,
                      pBlack[RGBA_RED],pBlack[RGBA_GREEN],pBlack[RGBA_BLUE],
                      255);
    }
    else
    {
      SetPaletteEntry(0,255,255,255,255);
      SetPaletteEntry(1,0,0,0,255);
    }

    int SrcLineLen = rSrcBmp.GetWidth();
    // assume msb is leftmost
    for (int y = 0; y<GetHeight(); ++y)
    { // For each line
      BYTE * pSrcPixel = rSrcBmp.GetLineArray()[y];
      BYTE * pDstPixel = GetLineArray()[y];
      for (int x = 0; x < SrcLineLen; ++x)  // For each pixel
        pDstPixel[x] = pSrcPixel[x / 8] & (128 >> (x & 7)) ? 1 : 0;
    }
  }
  const_cast<CBmp &>(rSrcBmp).Unlock();
  Unlock();
}

void CBmp::create1BPPCopy (const CBmp & rSrcBmp)
{
  int BPPSrc = rSrcBmp.GetBitsPerPixel();
  Lock(false, true);
  const_cast<CBmp &>(rSrcBmp).Lock(true, false);
  BYTE ** pSrcLines = rSrcBmp.GetLineArray();
  BYTE ** pDstLines = GetLineArray();
  int SrcLineLen = GetWidth();

  SetPaletteEntry(0,255,255,255,255);
  SetPaletteEntry(1,0,0,0,255);

  // downgrade to monochrome
  CPixel32 *pPal = rSrcBmp.GetPalette();
  BYTE *pRGBA;
  int DstLineLen = GetBytesPerLine();

  for (int y = 0; y < GetHeight(); ++y)
  { // For each line
    BYTE * pSrcPixel = pSrcLines[y];
    BYTE * pDstPixel = pDstLines[y];
    // fill with background (index 0) color
    memset(pDstPixel,0,DstLineLen);

    for (int x = 0; x < SrcLineLen; ++x)  // For each pixel
    {
      pRGBA = BPPSrc == 8 ? (BYTE*) &pPal[*pSrcPixel] : pSrcPixel;
      // the following criterion supposedly finds "dark" pixels; it may
      // need some twiddling and maybe use the alpha channel as well
      if (pRGBA[RGBA_RED] < 128 &&
          pRGBA[RGBA_GREEN] < 128 &&
          pRGBA[RGBA_BLUE] < 128 )
        pDstPixel[x / 8] |= 128 >> (x & 7);
      pSrcPixel += BPPSrc == 8 ? 1 : sizeof(CPixel32);
    }
  }
  const_cast<CBmp &>(rSrcBmp).Unlock();
  Unlock();
}

/*
/--------------------------------------------------------------------
|
|      $Log: bitmap.cpp,v $
|      Revision 1.27  2000/12/18 22:42:52  uzadow
|      Replaced RGBAPIXEL with CPixel32.
|
|      Revision 1.26  2000/12/03 21:15:43  uzadow
|      Fixed png handling in configure; test/ adapted to linux
|
|      Revision 1.25  2000/11/21 20:18:03  uzadow
|      Added operator ==
|
|      Revision 1.24  2000/11/02 21:28:47  uzadow
|      Fixed copy constructors.
|
|      Revision 1.23  2000/09/26 14:28:47  Administrator
|      Added Threshold filter
|
|      Revision 1.22  2000/09/26 12:14:51  Administrator
|      Refactored quantization.
|
|      Revision 1.21  2000/09/01 14:30:03  Administrator
|      no message
|
|      Revision 1.20  2000/09/01 14:19:46  Administrator
|      no message
|
|      Revision 1.19  2000/08/13 12:11:43  Administrator
|      Added experimental DirectDraw-Support
|
|      Revision 1.18  2000/05/28 11:08:51  Ulrich von Zadow
|      Corrected alpha channel bug in CreateCopy.
|
|      Revision 1.17  2000/03/31 12:20:05  Ulrich von Zadow
|      Video invert filter (beta)
|
|      Revision 1.16  2000/03/31 11:53:29  Ulrich von Zadow
|      Added quantization support.
|
|      Revision 1.15  2000/01/16 20:43:12  anonymous
|      Removed MFC dependencies
|
|      Revision 1.14  2000/01/10 23:52:59  Ulrich von Zadow
|      Changed formatting & removed tabs.
|
|      Revision 1.13  1999/12/31 17:59:54  Ulrich von Zadow
|      Corrected error in CBmp::SetPixel for 1 bpp.
|
|      Revision 1.12  1999/12/30 15:54:47  Ulrich von Zadow
|      Added CWinBmp::FromClipBoard() and CreateFromHBitmap().
|
|      Revision 1.11  1999/12/10 01:27:26  Ulrich von Zadow
|      Added assignment operator and copy constructor to
|      bitmap classes.
|
|      Revision 1.10  1999/12/08 15:39:45  Ulrich von Zadow
|      Unix compatibility changes
|
|      Revision 1.9  1999/12/02 17:07:34  Ulrich von Zadow
|      Changes by bdelmee.
|
|      Revision 1.8  1999/11/08 22:10:53  Ulrich von Zadow
|      no message
|
|      Revision 1.7  1999/10/22 21:25:51  Ulrich von Zadow
|      Removed buggy octree quantization
|
|      Revision 1.6  1999/10/21 18:47:26  Ulrich von Zadow
|      Added Rotate, GetPixel, SetPixel and FindNearestColor
|
|      Revision 1.5  1999/10/21 16:05:43  Ulrich von Zadow
|      Moved filters to separate directory. Added Crop, Grayscale and
|      GetAlpha filters.
|
|      Revision 1.4  1999/10/19 21:29:44  Ulrich von Zadow
|      Added filters.
|
|      Revision 1.3  1999/10/03 18:50:51  Ulrich von Zadow
|      Added automatic logging of changes.
|
|
\--------------------------------------------------------------------
*/

